import unittest
import binascii
from neo3 import vm, contracts
from neo3.core import syscall_name_to_int
from tests.contracts.interop.utils import test_engine


class StdLibTestCase(unittest.TestCase):
    def test_binary_serialization(self):
        engine = test_engine(has_snapshot=True)
        original_item = 100
        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call_with_args(contracts.StdLibContract().hash, "serialize", [original_item])
        engine.load_script(vm.Script(sb.to_array()))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        item = engine.result_stack.pop()
        self.assertIsInstance(item, vm.ByteStringStackItem)
        self.assertEqual(b'\x21\x01\x64', item.to_array())

    def test_binary_deserialization(self):
        engine = test_engine(has_snapshot=True)
        original_item = 100
        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call_with_args(contracts.StdLibContract().hash, "serialize", [original_item])
        # now take the results of the "serialize" call and call "deserialize" on the StdLib contract
        sb.emit_push(1)
        sb.emit(vm.OpCode.PACK)  # pack the results of "serialize" as the arguments for the next call
        sb.emit_push(0xF)  # CallFlags.ALL
        sb.emit_push("deserialize")
        sb.emit_push(contracts.StdLibContract().hash.to_array())
        sb.emit_syscall(syscall_name_to_int("System.Contract.Call"))
        engine.load_script(vm.Script(sb.to_array()))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        item = engine.result_stack.pop()
        self.assertEqual(original_item, int(item))

    def test_json_serialization(self):
        sb = vm.ScriptBuilder()
        sb.emit_push(5)
        sb.emit_push(1)  # arg len
        sb.emit(vm.OpCode.PACK)
        sb.emit_push(0xF)  # CallFlags.ALL
        sb.emit_push("jsonSerialize")
        sb.emit_push(contracts.StdLibContract().hash.to_array())
        sb.emit_syscall(syscall_name_to_int("System.Contract.Call"))

        sb.emit(vm.OpCode.PUSH0)
        sb.emit(vm.OpCode.NOT)
        sb.emit_push(1)  # arg len
        sb.emit(vm.OpCode.PACK)
        sb.emit_push(0xF)  # CallFlags.ALL
        sb.emit_push("jsonSerialize")
        sb.emit_push(contracts.StdLibContract().hash.to_array())
        sb.emit_syscall(syscall_name_to_int("System.Contract.Call"))

        sb.emit_push("test")
        sb.emit_push(1)  # arg len
        sb.emit(vm.OpCode.PACK)
        sb.emit_push(0xF)  # CallFlags.ALL
        sb.emit_push("jsonSerialize")
        sb.emit_push(contracts.StdLibContract().hash.to_array())
        sb.emit_syscall(syscall_name_to_int("System.Contract.Call"))

        sb.emit(vm.OpCode.PUSHNULL)
        sb.emit_push(1)  # arg len
        sb.emit(vm.OpCode.PACK)
        sb.emit_push(0xF)  # CallFlags.ALL
        sb.emit_push("jsonSerialize")
        sb.emit_push(contracts.StdLibContract().hash.to_array())
        sb.emit_syscall(syscall_name_to_int("System.Contract.Call"))

        sb.emit(vm.OpCode.NEWMAP)
        sb.emit(vm.OpCode.DUP)
        sb.emit_push("key")
        sb.emit_push("value")
        sb.emit(vm.OpCode.SETITEM)
        sb.emit_push(1)  # arg len
        sb.emit(vm.OpCode.PACK)
        sb.emit_push(0xF)  # CallFlags.ALL
        sb.emit_push("jsonSerialize")
        sb.emit_push(contracts.StdLibContract().hash.to_array())
        sb.emit_syscall(syscall_name_to_int("System.Contract.Call"))

        data = sb.to_array()

        engine = test_engine(has_snapshot=True)
        engine.load_script(vm.Script(data))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)

        def pop_to_human_readable():
            return binascii.unhexlify(str(engine.result_stack.pop()).encode()).decode()

        self.assertEqual('{"key":"value"}', pop_to_human_readable())
        self.assertEqual('null', pop_to_human_readable())
        self.assertEqual('"test"', pop_to_human_readable())
        self.assertEqual('true', pop_to_human_readable())
        self.assertEqual('5', pop_to_human_readable())

    def test_json_deserialization(self):
        script = vm.ScriptBuilder()
        script.emit_dynamic_call_with_args(contracts.StdLibContract().hash, "jsonDeserialize", ["{\"key\":\"value\"}"])
        script.emit_dynamic_call_with_args(contracts.StdLibContract().hash, "jsonDeserialize", ["null"])

        engine = test_engine(has_snapshot=True)
        data = script.to_array()
        engine.load_script(vm.Script(data))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        self.assertEqual(2, len(engine.result_stack._items))
        self.assertIsInstance(engine.result_stack.pop(), vm.NullStackItem)
        r = engine.result_stack.pop()
        self.assertIsInstance(r, vm.MapStackItem)
        self.assertEqual(1, len(r))
        self.assertEqual(vm.ByteStringStackItem("key"), r.keys()[0])
        self.assertEqual(vm.ByteStringStackItem("value"), r.values()[0])

    def test_atoi(self):
        engine = test_engine(has_snapshot=True)
        original_item = b'100'
        base = 10
        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call_with_args(contracts.StdLibContract().hash, "atoi", [original_item, base])
        engine.load_script(vm.Script(sb.to_array()))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        item = engine.result_stack.pop()
        self.assertEqual(vm.IntegerStackItem(100), item)

        engine = test_engine(has_snapshot=True)
        original_item = b'64'
        base = 16
        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call_with_args(contracts.StdLibContract().hash, "atoi", [original_item, base])
        engine.load_script(vm.Script(sb.to_array()))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        item = engine.result_stack.pop()
        self.assertEqual(vm.IntegerStackItem(100), item)

        engine = test_engine(has_snapshot=True)
        invalid_base = 2
        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call_with_args(contracts.StdLibContract().hash, "atoi", [original_item, invalid_base])
        engine.load_script(vm.Script(sb.to_array()))
        engine.execute()

        self.assertEqual(vm.VMState.FAULT, engine.state)
        self.assertIn("Invalid base specified", engine.exception_message)

    def test_itoa(self):
        engine = test_engine(has_snapshot=True)
        original_item = 100
        base = 10
        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call_with_args(contracts.StdLibContract().hash, "itoa", [original_item, base])
        engine.load_script(vm.Script(sb.to_array()))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        item = engine.result_stack.pop()
        self.assertEqual('100', item.to_array().decode('utf-8'))

        engine = test_engine(has_snapshot=True)
        base = 16
        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call_with_args(contracts.StdLibContract().hash, "itoa", [original_item, base])
        engine.load_script(vm.Script(sb.to_array()))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        item = engine.result_stack.pop()
        self.assertEqual('64', item.to_array().decode('utf-8'))

        engine = test_engine(has_snapshot=True)
        invalid_base = 2
        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call_with_args(contracts.StdLibContract().hash, "itoa", [original_item, invalid_base])
        engine.load_script(vm.Script(sb.to_array()))
        engine.execute()

        self.assertEqual(vm.VMState.FAULT, engine.state)
        self.assertIn("Invalid base specified", engine.exception_message)

    def test_base64_encode(self):
        engine = test_engine(has_snapshot=True)
        original_item = 100
        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call_with_args(contracts.StdLibContract().hash, "base64Encode", [original_item])
        engine.load_script(vm.Script(sb.to_array()))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        item = engine.result_stack.peek()
        self.assertEqual('ZA==', item.to_array().decode())

    def test_base64_decode(self):
        engine = test_engine(has_snapshot=True)
        original_item = 100
        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call_with_args(contracts.StdLibContract().hash, "base64Encode", [original_item])
        # now take the results of the "base64Encode" call and call "base64Decode" on the StdLib contract
        sb.emit_push(1)  # arg len
        sb.emit(vm.OpCode.PACK)  # pack the results of "base64Encode" as the arguments for the next call
        sb.emit_push(0xF)  # CallFlags.ALL
        sb.emit_push("base64Decode")
        sb.emit_push(contracts.StdLibContract().hash.to_array())
        sb.emit_syscall(syscall_name_to_int("System.Contract.Call"))
        engine.load_script(vm.Script(sb.to_array()))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        item = engine.result_stack.pop()
        self.assertEqual(original_item, int(item.to_biginteger()))

    def test_base58_encode(self):
        engine = test_engine(has_snapshot=True)
        original_item = 100
        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call_with_args(contracts.StdLibContract().hash, "base58Encode", [original_item])
        engine.load_script(vm.Script(sb.to_array()))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        item = engine.result_stack.peek()
        self.assertEqual('2j', item.to_array().decode())

    def test_base58_decode(self):
        engine = test_engine(has_snapshot=True)
        original_item = 100
        sb = vm.ScriptBuilder()
        sb.emit_dynamic_call_with_args(contracts.StdLibContract().hash, "base58Encode", [original_item])
        # now take the results of the "base58Encode" call and call "base64Decode" on the StdLib contract
        sb.emit_push(1)
        sb.emit(vm.OpCode.PACK)  # pack the results of "base58Encode" as the arguments for the next call
        sb.emit_push(0xF)  # CallFlags.ALL
        sb.emit_push("base58Decode")
        sb.emit_push(contracts.StdLibContract().hash.to_array())
        sb.emit_syscall(syscall_name_to_int("System.Contract.Call"))
        engine.load_script(vm.Script(sb.to_array()))
        engine.execute()
        self.assertEqual(vm.VMState.HALT, engine.state)
        item = engine.result_stack.pop()
        self.assertEqual(original_item, int(item.to_biginteger()))



import unittest
from neo3.core import serialization


class SerializableObj(serialization.ISerializable):
    """Helper class for tests"""
    def __init__(self, a: int = None):
        self.a = a

    def serialize(self, writer: serialization.BinaryWriter) -> None:
        writer.write_uint8(self.a)

    def deserialize(self, reader: serialization.BinaryReader) -> None:
        self.a = reader.read_uint8()

    def __len__(self):
        return 1


class ISerializableTestCase(unittest.TestCase):
    def test_deserialize_from_bytes(self):
        # test class method
        s1 = SerializableObj()
        obj = s1.deserialize_from_bytes(b'\x01')
        self.assertEqual(1, obj.a)


class BinaryReaderTestCase(unittest.TestCase):
    def test_read_bytes(self):
        with serialization.BinaryReader(b'\x01') as br:
            b = br.read_byte()
            self.assertEqual(b'\x01', b)

        with self.assertRaises(ValueError) as context:
            with serialization.BinaryReader(b'') as br:
                b = br.read_byte()
        self.assertIn("Could not read byte from empty stream", str(context.exception))


        input_data = b'\x01\x02\x03'
        with serialization.BinaryReader(input_data) as br:
            b = br.read_bytes(2)
            self.assertEqual(input_data[:2], b)

        # try reading more than available
        with self.assertRaises(ValueError) as context:
            with serialization.BinaryReader(input_data) as br:
                b = br.read_bytes(5)
        self.assertIn("Could not read 5 bytes from stream. Only found 3 bytes of data", str(context.exception))

    def test_read_bool(self):
        with serialization.BinaryReader(b'\x00') as br:
            b = br.read_bool()
            self.assertIsInstance(b, bool)
            self.assertFalse(b)

        with serialization.BinaryReader(b'\x01') as br:
            b = br.read_bool()
            self.assertIsInstance(b, bool)
            self.assertTrue(b)

        with serialization.BinaryReader(b'\x55') as br:
            b = br.read_bool()
            self.assertIsInstance(b, bool)
            self.assertTrue(b)

    def test_reading_uint8(self):
        input_data = b'\x01\x02'
        # make sure we read only 1 byte
        with serialization.BinaryReader(input_data) as br:
            b = br.read_uint8()
            self.assertEqual(input_data[0], b)

        # validate we read an unsigned byte
        with serialization.BinaryReader(b'\xFF') as br:
            b = br.read_uint8()
            self.assertEqual(255, b)

    def test_read_uint16(self):
        input_data = b'\x01\x02\x03'
        # make sure we read only 2 bytes
        with serialization.BinaryReader(input_data) as br:
            b = br.read_uint16()
            self.assertEqual(int.from_bytes(input_data[:2],'little'), b)

        # validate we read as unsigned
        with serialization.BinaryReader(b'\x01\xFF') as br:
            b = br.read_uint16()
            self.assertEqual(65281, b)

    def test_read_int16(self):
        input_data = b'\x01\xFF'
        with serialization.BinaryReader(input_data) as br:
            b = br.read_int16()
            self.assertEqual(-255, b)

    def test_read_uint32(self):
        input_data = b'\x01\x02\x03\x04\x05'
        # make sure we read only 4 bytes
        with serialization.BinaryReader(input_data) as br:
            b = br.read_uint32()
            self.assertEqual(int.from_bytes(input_data[:4],'little'), b)

        # validate we read as unsigned
        with serialization.BinaryReader(b'\x01\x02\x03\xFF') as br:
            b = br.read_uint32()
            self.assertEqual(4278387201, b)

    def test_read_int32(self):
        with serialization.BinaryReader(b'\x01\x02\x03\xFF') as br:
            b = br.read_int32()
            self.assertEqual(-16580095, b)

    def test_read_uint64(self):
        input_data = b'\x01\x02\x03\x04\x05\x06\x07\x08\x09'
        # make sure we read only 4 bytes
        with serialization.BinaryReader(input_data) as br:
            b = br.read_uint64()
            self.assertEqual(int.from_bytes(input_data[:8],'little'), b)

        # validate we read as unsigned
        with serialization.BinaryReader(b'\x01\x02\x03\x04\x05\x06\x07\xFF') as br:
            b = br.read_uint64()
            self.assertEqual(18376663423120507393, b)

    def test_read_int64(self):
        with serialization.BinaryReader(b'\x01\x02\x03\x04\x05\x06\x07\xFF') as br:
            b = br.read_int64()
            self.assertEqual(-70080650589044223, b)

    def test_read_var_int(self):
        # no value
        with serialization.BinaryReader(b'\x00') as br:
            b = br.read_var_int()
            self.assertEqual(0, b)

        # a value smaller than 0xFD is encoded in 1 byte
        with serialization.BinaryReader(b'\xFC') as br:
            b = br.read_var_int()
            self.assertEqual(252, b)

        # a value smaller than 0xFFFF is encoded in 3 bytes
        input_data = b'\xfd\x01\xFF'
        with serialization.BinaryReader(input_data) as br:
            b = br.read_var_int()
            self.assertEqual(65281, b)

        # a value smaller than 0xFFFFFFFF is encoded in 5 bytes
        input_data = b'\xfe\x01\x02\x03\xFF'
        with serialization.BinaryReader(input_data) as br:
            b = br.read_var_int()
            self.assertEqual(4278387201, b)

        # a value bigger than 0xFFFFFFFF is encoded in 9 bytes
        input_data = b'\xff\x01\x02\x03\x04\x05\x06\x07\x08\xFF'
        with serialization.BinaryReader(input_data) as br:
            b = br.read_var_int()
            self.assertEqual(578437695752307201, b)

        # test reader with max size
        with self.assertRaises(ValueError) as context:
            with serialization.BinaryReader(b'\xFC') as br:
                b = br.read_var_int(max=10)
        self.assertIn("Invalid format", str(context.exception))

    def test_read_var_bytes(self):
        input_data = b'\x02\x01\x02\x03'
        with serialization.BinaryReader(input_data) as br:
            b = br.read_var_bytes()
            self.assertEqual(input_data[1:3], b)

        """
            test reading with insufficient data
            the C# equivalent will try to read up to what's encoded amount of bytes from the stream or less!
            
            byte[] value = { 0x2 };
            using (MemoryStream ms = new MemoryStream(value, false))
            using (BinaryReader reader = new BinaryReader(ms, System.Text.Encoding.UTF8))
            {
                byte[] b = reader.ReadVarBytes();
            }

        # encode 2 bytes of remaining data, but supply none
        """
        input_data = b'\x02'
        with serialization.BinaryReader(input_data) as br:
            b = br.read_var_bytes()
            self.assertEqual(b'', b)

    def test_read_string(self):
        input_data = b'\x02\x41\x42'
        with serialization.BinaryReader(input_data) as br:
            s = br.read_var_string()
            self.assertIsInstance(s, str)
            self.assertEqual('AB', s)

        # test with insufficient data
        input_data = b'\x02\x41'
        with self.assertRaises(ValueError) as context:
            with serialization.BinaryReader(input_data) as br:
                s = br.read_var_string()
        self.assertIn("unpack requires a buffer of 2 bytes", str(context.exception))

    def test_reading_serializable(self):
        s1 = SerializableObj(1)

        with serialization.BinaryReader(s1.to_array()) as br:
            obj = br.read_serializable(obj_type=SerializableObj)
            self.assertIsInstance(obj, serialization.ISerializable)
            self.assertIsInstance(obj, SerializableObj)
            self.assertEqual(1, obj.a)

    def test_reading_list_of_serializables(self):
        s1 = SerializableObj(1)
        s2 = SerializableObj(2)

        array_length = b'\x02'
        with serialization.BinaryReader(array_length + s1.to_array() + s2.to_array()) as br:
            objs = br.read_serializable_list(obj_type=SerializableObj)
            self.assertIsInstance(objs, list)
            self.assertTrue(2, len(objs))
            for o in objs:
                self.assertIsInstance(o, SerializableObj)

        # test but limit max array
        array_length = b'\x02'
        with serialization.BinaryReader(array_length + s1.to_array() + s2.to_array()) as br:
            objs = br.read_serializable_list(obj_type=SerializableObj, max=1)
            self.assertIsInstance(objs, list)
            self.assertTrue(1, len(objs))
            for o in objs:
                self.assertIsInstance(o, SerializableObj)

    def test_length(self):
        input_data = b'\x02\x41\x42'
        with serialization.BinaryReader(input_data) as br:
            self.assertEqual(3, len(br))

class BinaryWriterTestCase(unittest.TestCase):
    def test_write_bytes(self):
        with serialization.BinaryWriter() as bw:
            bw.write_uint8(5)
            bw.write_uint16(257)
            self.assertEqual(b'\x05\x01\x01', bw._stream.getvalue())

    def test_write_bool(self):
        with serialization.BinaryWriter() as bw:
            bw.write_bool(0)
            bw.write_bool(1)
            bw.write_bool(15)
            self.assertEqual(b'\x00\x01\x01', bw._stream.getvalue())

    def test_write_uint8(self):
        with serialization.BinaryWriter() as bw:
            bw.write_uint8(255)
            # this also validates signed vs unsigned. If it was signed it would need an extra \x00 to express the value
            # and would not fit in 1 byte
            self.assertEqual(b'\xFF', bw._stream.getvalue())

    def test_write_uint16(self):
        with serialization.BinaryWriter() as bw:
            bw.write_uint16(0xFFFF)
            # this also validates signed vs unsigned. If it was signed it would need an extra \x00 to express the value
            # and would not fit in 2 bytes
            self.assertEqual(b'\xFF\xFF', bw._stream.getvalue())

    def test_write_uint32(self):
        with serialization.BinaryWriter() as bw:
            bw.write_uint32(0xFFFFFFFF)
            # this also validates signed vs unsigned. If it was signed it would need an extra \x00 to express the value
            # and would not fit in 4 bytes
            self.assertEqual(b'\xFF\xFF\xFF\xFF', bw._stream.getvalue())

    def test_write_uint64(self):
        with serialization.BinaryWriter() as bw:
            bw.write_uint64(0xFFFFFFFFFFFFFFFF)
            # this also validates signed vs unsigned. If it was signed it would need an extra \x00 to express the value
            # and would not fit in 8 bytes
            self.assertEqual(b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF', bw._stream.getvalue())

    def test_write_int16(self):
        with serialization.BinaryWriter() as bw:
            bw.write_int16(-1)
            # this also validates signed vs unsigned. If it was unsigned it would be without \x00
            self.assertEqual(b'\xFF\xFF', bw._stream.getvalue())

    def test_write_int32(self):
        with serialization.BinaryWriter() as bw:
            bw.write_int32(-1)
            # this also validates signed vs unsigned. If it was unsigned it would be without \x00
            self.assertEqual(b'\xFF\xFF\xFF\xFF', bw._stream.getvalue())

    def test_write_int64(self):
        with serialization.BinaryWriter() as bw:
            bw.write_int64(-1)
            # this also validates signed vs unsigned. If it was unsigned it would be without \x00
            self.assertEqual(b'\xFF\xFF\xFF\xFF\xFF\xFF\xFF\xFF', bw._stream.getvalue())

    def test_write_var_string(self):
        with serialization.BinaryWriter() as bw:
            bw.write_var_string('ABC')
            self.assertEqual(b'\x03\x41\x42\x43', bw._stream.getvalue())

    def test_write_var_int(self):
        with self.assertRaises(TypeError) as context:
            with serialization.BinaryWriter() as bw:
                bw.write_var_int(b'\x01')
        self.assertIn("not int type.", str(context.exception))

        with self.assertRaises(ValueError) as context:
            with serialization.BinaryWriter() as bw:
                bw.write_var_int(-1)
        self.assertIn("too small.", str(context.exception))

        with serialization.BinaryWriter() as bw:
            bw.write_var_int(1)
            self.assertEqual(b'\x01', bw._stream.getvalue())

        with serialization.BinaryWriter() as bw:
            bw.write_var_int(65535)  # 0xFFFF edge
            self.assertEqual(b'\xfd\xff\xFF', bw._stream.getvalue())

        with serialization.BinaryWriter() as bw:
            bw.write_var_int(4294967295)  # 0xFFFFFFFF edge
            self.assertEqual(b'\xfe\xff\xff\xff\xff', bw._stream.getvalue())

        with serialization.BinaryWriter() as bw:
            bw.write_var_int(4294967296)
            self.assertEqual(b'\xff\x00\x00\x00\x00\x01\x00\x00\x00', bw._stream.getvalue())

    def test_write_var_bytes(self):
        with serialization.BinaryWriter() as bw:
            bw.write_var_bytes(b'\x01\x02\x03\x04')
            self.assertEqual(b'\x04\x01\x02\x03\x04', bw._stream.getvalue())

    def test_write_serializable(self):
        s1 = SerializableObj(1)
        with serialization.BinaryWriter() as bw:
            bw.write_serializable(s1)
            self.assertEqual(b'\x01', bw._stream.getvalue())

    def test_write_list_of_serializable_objects(self):
        s1 = SerializableObj(1)
        s2 = SerializableObj(3)
        with serialization.BinaryWriter() as bw:
            bw.write_serializable_list([s1, s2])
            self.assertEqual(b'\x02\x01\x03', bw._stream.getvalue())

    def test_length(self):
        with serialization.BinaryWriter() as br:
            br.write_uint16(1000)
            self.assertEqual(2, len(br))

        with serialization.BinaryWriter() as br:
            br.write_uint64(1000)
            self.assertEqual(8, len(br))




from __future__ import annotations
from typing import Optional
from neo3 import contracts, storage
from neo3.contracts.interop import register, IIterator, StorageIterator

MAX_STORAGE_KEY_SIZE = 64
MAX_STORAGE_VALUE_SIZE = 65535


@register("System.Storage.GetContext", 1 << 4, contracts.CallFlags.READ_STATES)
def get_context(engine: contracts.ApplicationEngine) -> storage.StorageContext:
    contract = contracts.ManagementContract().get_contract(engine.snapshot, engine.current_scripthash)
    if contract is None:
        raise ValueError("Contract not deployed")
    return storage.StorageContext(contract.id, False)


@register("System.Storage.GetReadOnlyContext", 1 << 4, contracts.CallFlags.READ_STATES)
def get_read_only_context(engine: contracts.ApplicationEngine) -> storage.StorageContext:
    contract = contracts.ManagementContract().get_contract(engine.snapshot, engine.current_scripthash)
    if contract is None:
        raise ValueError("Contract not deployed")
    return storage.StorageContext(contract.id, True)


@register("System.Storage.AsReadOnly", 1 << 4, contracts.CallFlags.READ_STATES)
def context_as_read_only(engine: contracts.ApplicationEngine,
                         context: storage.StorageContext) -> storage.StorageContext:
    if not context.is_read_only:
        context = storage.StorageContext(context.id, True)
    return context


@register("System.Storage.Get", 1 << 15, contracts.CallFlags.READ_STATES)
def storage_get(engine: contracts.ApplicationEngine, context: storage.StorageContext, key: bytes) -> Optional[bytes]:
    storage_key = storage.StorageKey(context.id, key)
    item = engine.snapshot.storages.try_get(storage_key, read_only=True)
    if item is not None:
        return item.value
    return None


@register("System.Storage.Find", 1 << 15, contracts.CallFlags.READ_STATES)
def storage_find(engine: contracts.ApplicationEngine,
                 context: storage.StorageContext,
                 key: bytes,
                 options: contracts.FindOptions) -> IIterator:
    opt = contracts.FindOptions
    if opt.KEYS_ONLY in options and (
            opt.VALUES_ONLY in options
            or opt.DESERIALIZE_VALUES in options
            or opt.PICK_FIELD0 in options
            or opt.PICK_FIELD1 in options):
        raise ValueError("KEYS_ONLY and (VALUES_ONLY || DESERIALIZE_VALUES || PICK_FIELD0 || PICK_FIELD1) are mutually "
                         "exclusive")
    if opt.VALUES_ONLY in options and (opt.KEYS_ONLY in options or opt.REMOVE_PREFIX in options):
        raise ValueError("VALUES_ONLY and (KEYS_ONLY || REMOVE_PREFIX) are mutually exclusive")
    if opt.PICK_FIELD0 in options and opt.PICK_FIELD1 in options:
        raise ValueError("PICK_FIELD0 and PICK_FIELD1 are mutually exclusive")
    if (opt.PICK_FIELD0 in options or opt.PICK_FIELD1 in options) and opt.DESERIALIZE_VALUES not in options:
        raise ValueError("Can't use PICK_FIELD option without DESERIALIZE_VALUES")

    prefix_key = context.id.to_bytes(4, 'little', signed=True) + key
    it = StorageIterator(engine.snapshot.storages.find(prefix_key), len(key), options, engine.reference_counter)
    return it


@register("System.Storage.Put", 1 << 15, contracts.CallFlags.WRITE_STATES)
def storage_put(engine: contracts.ApplicationEngine, context: storage.StorageContext, key: bytes, value: bytes) -> None:
    if len(key) > MAX_STORAGE_KEY_SIZE:
        raise ValueError(f"Storage key length exceeds maximum of {MAX_STORAGE_KEY_SIZE}")
    if len(value) > MAX_STORAGE_VALUE_SIZE:
        raise ValueError(f"Storage value length exceeds maximum of {MAX_STORAGE_VALUE_SIZE}")
    if context.is_read_only:
        raise ValueError("Cannot persist to read-only storage context")

    storage_key = storage.StorageKey(context.id, key)
    item = engine.snapshot.storages.try_get(storage_key, read_only=False)

    if item is None:
        new_data_len = len(key) + len(value)
        item = storage.StorageItem(b'')
        engine.snapshot.storages.put(storage_key, item)
    else:
        if len(value) == 0:
            new_data_len = 0
        elif len(value) <= len(item.value):
            new_data_len = (len(value) - 1) // 4 + 1
        elif len(item.value) == 0:
            new_data_len = len(value)
        else:
            new_data_len = (len(item.value) - 1) // 4 + 1 + len(value) - len(item.value)

    engine.add_gas(new_data_len * engine.STORAGE_PRICE)
    item.value = value


@register("System.Storage.Delete", 1 << 15, contracts.CallFlags.WRITE_STATES)
def storage_delete(engine: contracts.ApplicationEngine, context: storage.StorageContext, key: bytes) -> None:
    if context.is_read_only:
        raise ValueError("Cannot delete from read-only storage context")
    storage_key = storage.StorageKey(context.id, key)
    engine.snapshot.storages.delete(storage_key)
